from abc import ABC
# from engine.exceptions import ValidationError


class BaseRule(ABC):
    # errors = None
    name = "base_rules.default"
    message = "Blank Row Found"
    sevearity = ''
    
    @property
    def code(self):
        if self.name:
            return self.name
        else:
            self.__class__.__name__
    
    @property
    def errors(self):
        return {
            "sevearity": self.sevearity,
            "category": self.code,
            "reason": self.message
        }
    
    def is_valid(self, item):
        if not len(item):
            return False
        return True



from engine.data import reader, writer
from engine.exceptions import ImproperlyConfigured
from engine.conf.settings import VALIDATION_PIPELINES
from engine.validations.base_rules import *

__all__ = [
    "Engine"
 ]


class BaseValidationEngine(object):
    validation_rules = VALIDATION_PIPELINES
    data_reader = None
    data_writer = None
    
    def __init__(self, 
        input_file_path, 
        output_file_path,
        S3_config=None
        ) -> None:
        
        self.input_file_path = input_file_path
        self.output_file_path = output_file_path
        self.S3_config = S3_config or {}
        
        self.items = []

    def all(self):
        if self.data_reader is None:
            raise ImproperlyConfigured(
                "Invalid Data Reader class !"
            )
        rows = []
        
        for row in self.data_reader(
                    file_path=self.input_file_path,
                    **self.S3_config
                ).read():
            rows.append(row)
        return rows
    
    def __import_model(self, name:str):
        components = name.split('.')
        module = __import__(components[0])
        for comp in components[1:]:
            module = getattr(module, comp)        
        return module
        
    
    def validate(self, row):
        for validation_cls in self.validation_rules:
            klass = self.__import_model(validation_cls)
            validation = klass()
            if not validation.is_valid(row):
                return False, validation.errors
        return True, {"category": "success", "sevearity": "",  "reason": "success"}
    
    def validate_all(self):
        for item in self.all():
            is_valid, response = self.validate(item)
            print(response)
            self.items.append({**item, 'is_valid': is_valid, **response})
        return self.items
    
    def save(self):
        if self.data_writer is None:
            raise ImproperlyConfigured(
                "Invalid Data Writer class"
            )  
        if len(self.items):
            fieldnames = list(self.items[0].keys())
            print(fieldnames)
            self.data_writer(
                file_path=self.output_file_path,
                **self.S3_config
            ).write(self.items, headers=fieldnames)
        
    
class ValidationEngine(BaseValidationEngine):
    data_reader = reader.CSVReader
    data_writer = writer.CSVWriter
    

Engine = ValidationEngine


from engine.core import Engine

if __name__ == "__main__":
    
    engine = Engine(
        input_file_path="input.csv",
        output_file_path="output.csv"
    )
    engine.validate_all()
    engine.save()





# inlude validation pipeline here

# EMPTY VALUES configuration

EMPTY_VALUES = (None, "", [], (), {})

# General Validation Out of range configuration

GEN_VAL_OUT_OF_RANGE_VALUE = {
  'min_value': 100000.0,
  'max_value': 10000000.0
}

# Listing price revisioning validation rule delta percentage from last Day configuration
# value range from +- 0 to 100 
# data type float
LP_VAL_LAST_DAY_DELTA_PCT__ALLOWED_RANGE  = [-5.0, 5.0]


# Validation Pipelies
VALIDATION_PIPELINES = [
  'engine.validations.general_rules.NullNegativeZeroValidationRule',
  'engine.validations.general_rules.OutOfRangeValidationRule',
  'engine.validations.revision_rules.DeltaPercentageFromLastDayRule',
]




import csv
from io import StringIO
from engine.exceptions import ImproperlyConfigured
from engine.conf import settings
import boto3


class BaseReader(object):
    
    def read(self):
        raise NotImplemented("")


class CSVReader(BaseReader):
    
    def __init__(self, file_path, *args, **kwargs) -> None:
        print(file_path)
        self.file_path = file_path
        self.__args = args
        self.__kwrgs = kwargs
    
    def get_fs(self):
        return open(self.file_path)
            
    def read(self):
        fs = self.get_fs()
        for row in csv.DictReader(fs, *self.__args, **self.__kwrgs):
            yield row
        fs.close()
                    
    
class S3CSVReader(CSVReader):
    
    def __init__(self, AWS_ACCESS_KEY_ID, AWS_SECRET_KEY, bucket_name, file_key, *args, **kwargs) -> None:
        super().__init__(file_path=file_key, *args, **kwargs)
        
        self.BUCKET = bucket_name
        self.AWS_ACCESS_KEY_ID = AWS_ACCESS_KEY_ID
        self.AWS_SECRET_KEY = AWS_SECRET_KEY
        self.client = boto3.client('s3', aws_access_key_id=self.AWS_ACCESS_KEY_ID, aws_secret_access_key=self.AWS_SECRET_KEY)
        
    def get_fs(self):
        file_object = self.client.get_object(Bucket=self.BUCKET, key=self.file_path)
        return StringIO(file_object['Body'].read().decode('utf-8'))
    
    
class GCSCSVReader(CSVReader):
    
    def __init__(self, AWS_ACCESS_KEY_ID, AWS_SECRET_KEY, bucket_name, file_key, *args, **kwargs) -> None:
        super().__init__(file_path=file_key, *args, **kwargs)
        
        self.BUCKET = bucket_name
        self.AWS_ACCESS_KEY_ID = AWS_ACCESS_KEY_ID
        self.AWS_SECRET_KEY = AWS_SECRET_KEY
        self.client = boto3.client('s3', aws_access_key_id=self.AWS_ACCESS_KEY_ID, aws_secret_access_key=self.AWS_SECRET_KEY)
        
    def get_fs(self):
        file_object = self.client.get_object(Bucket=self.BUCKET, key=self.file_path)
        return StringIO(file_object['Body'].read().decode('utf-8'))
    
    
    


from engine.validations.base_rules import BaseRule


class DeltaPercentageFromLastDayRule(BaseRule):
    name = "last-price-revision-delta-percentage-rule"
    sevearity = None
    message = "Delta Range greater than +-5% from Last Day. Actual is {1}%"
    
    def is_valid(self, item, target_field="final_liquidation_price"):
        final_lp = float(item['final_liquidation_price'])
        lp = float(item['lp'])
        
        delta = final_lp - lp
        delta_pct = round(100 * delta / lp ,0)
        
        if  -5 <= delta_pct  <= 5:
            return True
        
        elif  -15 <= delta_pct  <= 15:
            self.sevearity = 'LOW'
            self.message= self.message.format("15%", delta_pct) 
            return False
        
        elif  -20 <= delta_pct  <= 20:
            self.sevearity = 'MEDIUM'
            self.message = self.message.format("20%", delta_pct) 
            return False
        
        else:
            self.sevearity = 'HIGH'
            self.message= "Delta is not allowed more than 20%. Actual is {}".format(delta_pct) 
            return False    
    
    
class UnknownRule(BaseRule):
    name = "unknown-rule"
    sevearity = None
    message = "Delta Range greater than +-5% from Last Day. Actual is {1}%"
    
    def is_valid(self, item, target_field="final_liquidation_price"):
        final_lp = float(item['final_liquidation_price'])
        lp = float(item['lp'])
        
        delta = final_lp - lp
        delta_pct = round(100 * delta / lp ,0)
        
        if  -5 <= delta_pct  <= 5:
            return True
        
        elif  -15 <= delta_pct  <= 15:
            self.sevearity = 'LOW'
            self.message= self.message.format("15%", delta_pct) 
            return False
        
        elif  -20 <= delta_pct  <= 20:
            self.sevearity = 'MEDIUM'
            self.message = self.message.format("20%", delta_pct) 
            return False
        
        else:
            self.sevearity = 'HIGH'
            self.message= "Delta is not allowed more than 20%. Actual is {}".format(delta_pct) 
            return False    
    


from setuptools import find_packages, setup

with open('README.md') as f:
    README = f.read()
    
setup(
    name='price-val-engine',
    packages=find_packages(),
    version='0.1.0',
    description='last price revision validation rules',
    # long_description=README,
    author='Chandan Kumar Ojha',
    author_email="mr.chandanojha@gmail.com",
    license='MIT',
    python_requires=">=3.0",
    install_requires=[
        'boto3==1.21.45',
        'botocore==1.24.45'
    ]
)


import operator

NON_FIELD_ERRORS = "__all__"


def is_iterable(x):
    "An implementation independent way of checking for iterables"
    try:
        iter(x)
    except TypeError:
        return False
    else:
        return True
    
def make_hashable(value):
    """
    Attempt to make value hashable or raise a TypeError if it fails.

    The returned value should generate the same hash for equal values.
    """
    if isinstance(value, dict):
        return tuple([
            (key, make_hashable(nested_value))
            for key, nested_value in sorted(value.items())
        ])
    # Try hash to avoid converting a hashable iterable (e.g. string, frozenset)
    # to a tuple.
    try:
        hash(value)
    except TypeError:
        if is_iterable(value):
            return tuple(map(make_hashable, value))
        # Non-hashable, non-iterable.
        raise
    return value

class ImproperlyConfigured(Exception):
    """Data is somehow improperly configured"""

    pass

class ValidationError(Exception):
    """An error while validating data."""

    def __init__(self, message, code=None, params=None):
        """
        The `message` argument can be a single error, a list of errors, or a
        dictionary that maps field names to lists of errors. What we define as
        an "error" can be either a simple string or an instance of
        ValidationError with its message attribute set, and what we define as
        list or dictionary can be an actual `list` or `dict` or an instance
        of ValidationError with its `error_list` or `error_dict` attribute set.
        """
        super().__init__(message, code, params)

        if isinstance(message, ValidationError):
            if hasattr(message, "error_dict"):
                message = message.error_dict
            elif not hasattr(message, "message"):
                message = message.error_list
            else:
                message, code, params = message.message, message.code, message.params

        if isinstance(message, dict):
            self.error_dict = {}
            for field, messages in message.items():
                if not isinstance(messages, ValidationError):
                    messages = ValidationError(messages)
                self.error_dict[field] = messages.error_list

        elif isinstance(message, list):
            self.error_list = []
            for message in message:
                # Normalize plain strings to instances of ValidationError.
                if not isinstance(message, ValidationError):
                    message = ValidationError(message)
                if hasattr(message, "error_dict"):
                    self.error_list.extend(sum(message.error_dict.values(), []))
                else:
                    self.error_list.extend(message.error_list)

        else:
            self.message = message
            self.code = code
            self.params = params
            self.error_list = [self]

    @property
    def message_dict(self):
        # Trigger an AttributeError if this ValidationError
        # doesn't have an error_dict.
        getattr(self, "error_dict")

        return dict(self)

    @property
    def messages(self):
        if hasattr(self, "error_dict"):
            return sum(dict(self).values(), [])
        return list(self)

    def update_error_dict(self, error_dict):
        if hasattr(self, "error_dict"):
            for field, error_list in self.error_dict.items():
                error_dict.setdefault(field, []).extend(error_list)
        else:
            error_dict.setdefault(NON_FIELD_ERRORS, []).extend(self.error_list)
        return error_dict

    def __iter__(self):
        if hasattr(self, "error_dict"):
            for field, errors in self.error_dict.items():
                yield field, list(ValidationError(errors))
        else:
            for error in self.error_list:
                message = error.message
                if error.params:
                    message %= error.params
                yield str(message)

    def __str__(self):
        if hasattr(self, "error_dict"):
            return repr(dict(self))
        return repr(list(self))

    def __repr__(self):
        return "ValidationError(%s)" % self

    def __eq__(self, other):
        if not isinstance(other, ValidationError):
            return NotImplemented
        return hash(self) == hash(other)

    def __hash__(self):
        if hasattr(self, "message"):
            return hash(
                (
                    self.message,
                    self.code,
                    make_hashable(self.params),
                )
            )
        if hasattr(self, "error_dict"):
            return hash(make_hashable(self.error_dict))
        return hash(tuple(sorted(self.error_list, key=operator.attrgetter("message"))))
    
    


from engine import utils
from engine.conf import settings
from engine.validations.base_rules import BaseRule


EMPTY_VALUES = settings.EMPTY_VALUES or (None, "", [], (), {})

class NullNegativeZeroValidationRule(BaseRule):
    name = 'null-negative-zero-rule'
    message = "Invalid Value"
    sevearity = 'HIGH'
    
    def is_valid(self, item, target_field="final_liquidation_price"):
        print(item)
        value = item[target_field]
        if value in EMPTY_VALUES:
            self.message = "Null Value"
            return False
        if not utils.is_number(value):
            self.message = "Number value Expected"
            return False
        if float(value) == 0.0:
            self.message = "Zero Value"
            return False
        if float(value) < 0.0:
            self.message = "Negative Value"
            return False
        return True

class OutOfRangeValidationRule(BaseRule):
    name = "out-of-range-rule"
    message = "Out of Range"
    sevearity = 'HIGH'
    
    min_value = settings.GEN_VAL_OUT_OF_RANGE_VALUE.get('min_value') or  100000.0
    max_value = settings.GEN_VAL_OUT_OF_RANGE_VALUE.get('max_value') or 10000000.0
    
    def cal_min_value(self):
        return 100
    
    
    
    def is_valid(self, item, target_field="final_liquidation_price"):
        value = item[target_field]
        min_value =  self.cal_min_value()
        if self.min_value <= float(value) > self.max_value:
            self.message = f"{target_field} {value} out of range. <min{self.min_value}> - <max{self.max_value}>"
            return False
        
        return True

